!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: indexlist.gs,v 1.7 2008-01-09 19:02:34 stever Exp $
!
! Superclass Hierarchy:
!   IndexList, Array, SequenceableCollection, Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods IndexList
removeallclassmethods IndexList
protectmethods

category: 'For Documentation Installation only'
classmethod: IndexList
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Class methods for IndexList

category: 'Instance Creation'
classmethod: IndexList
new

"Returns a new, initialized instance of the receiver."
<protected>

^ super new initialize
%

! ------------------- Instance methods for IndexList

category: 'Searching'
method: IndexList
_findOffsetFor: indexObj withOffset: anInteger

"Returns the integer entry number (position in the index list) at which
 indexObj is located in the receiver; or nil if it is not found."
<protected>

"Performs a linear search."

1 to: self size by: 2 do: [ :i |
  ( indexObj == (self at: i) _and: [ (self at: i + 1) == anInteger ] )
        ifTrue: [ ^ i ]
].
^ nil
%

category: 'Searching'
method: IndexList
_findOffsetFor: indexObj

"Returns the integer entry number (position in the index list) at which
 indexObj is located in the receiver; or nil if it is not found."

<protected>

"Performs a linear search."

1 to: self size by: 2 do: [ :i |
  indexObj == (self at: i)
        ifTrue: [ ^ i ]
].
^ nil
%

category: 'Updating Indexes'
method: IndexList
_putCommonPathTermsForPathArray: pathArray into: indexObj

"Find the common path terms that already exist for indexes whose path
 components match pathArray (an Array of Strings).  Adds those path terms to
 the end of given indexObj.  Do not put a common path term into indexObj that
 is the last path term for an existing index unless the existing index
 is an identity index."
<protected>

| firstUnique somePathTerms j minSize sz someSz someIsIdentityIndex isRangeIndex |
firstUnique := 0.
isRangeIndex := indexObj isRangeEqualityIndex.
sz := pathArray size.
1 to: self size by: 2 do: [ :i |
    j := self at: i + 1.
    " only consider index objects for which the NSC is the root "
    j == 1
        ifTrue: [
            somePathTerms := self at: i.
            someIsIdentityIndex := somePathTerms isIdentityIndex.
            someSz := somePathTerms size.
            minSize := someSz min: sz.

            [ j <= minSize _and:
            [ (somePathTerms at: j) name = (pathArray at: j) ] ] whileTrue: [
                ( j > firstUnique _and:
                [ ( j < someSz _or: [ someIsIdentityIndex _and: [ j == someSz ] ] )_and:
                [ j ~= sz _or: [ isRangeIndex not ] ] ] )
                   ifTrue: [
                       firstUnique := j.
                       indexObj addLast: (somePathTerms at: j)
                   ].
                j := j + 1
            ]
        ]
].
%

category: 'Updating'
method: IndexList
addIndex: anIndexObj withOffset: anInteger nsc: anNsc

"Add an entry (index object, path offset) to the end of the index list."
<protected>

| rootTerm |
" do not allow duplicate entries "
1 to: self size by: 2 do: [ :i |
    " if entry already exists, just returns "
  anIndexObj == (self at: i)
    ifTrue: [
      (self at: i + 1) == anInteger
        ifTrue: [ ^ self ]
        ifFalse: [
          anNsc _error: #rtErrNscParticipatesInMultipleTerms
            args: #[ anIndexObj pathComponentsString ]
        ]
    ]
].

anIndexObj size < anInteger
    ifTrue: [ rootTerm := anIndexObj lastPathTerm ]
    ifFalse: [ rootTerm := anIndexObj at: anInteger ].

(rootTerms includesIdentical: rootTerm)
    ifTrue: [ rootTerm := nil ].

" add the entry to the end of the index list "
self _addIndex: anIndexObj withOffset: anInteger rootTerm: rootTerm
%

category: 'Updating'
method: IndexList
_addIndex: anIndexObj withOffset: anInteger rootTerm: rootTerm

"Add an entry (index object, path offset) to the end of the index list
 and update the root terms if necessary."
<protected>

self addLast: anIndexObj.
self addLast: anInteger.

rootTerm ~~ nil
    ifTrue: [ rootTerms addLast: rootTerm ].
%

category: 'Updating Indexes'
method: IndexList
buildPathTermsFor: indexObj with: pathArray

"Add path terms to the given indexObj, using the pathArray (Array of Strings).
 First find all common path terms (from existing indexes), then create any new
 path terms."
<protected>

| pathTerm startOffset constraintArray ivOffset |
" set the common path terms "
self _putCommonPathTermsForPathArray: pathArray into: indexObj.

" keep the starting point so we don't add duplicate entries "
startOffset := indexObj size + 1.

" add the index to shared path terms "
1 to: indexObj size do: [ :i |
    (indexObj at: i) addLast: indexObj
].

constraintArray := indexObj nscRoot _getConstraintsOnPath: pathArray.

" create any new path terms that are needed "
startOffset to: pathArray size do: [ :i |
    (pathArray at: i) = #* 
        ifTrue: [ pathTerm := SetValuedPathTerm new ]
        ifFalse: [
            (constraintArray at: i) == nil
                ifTrue: [ pathTerm := PathTerm new ]
                ifFalse: [
                    ivOffset := (constraintArray at: i) _ivOffsetOf: (pathArray at: i).
                    ivOffset == nil
                        ifTrue: [ pathTerm := PathTerm new ]
                        ifFalse: [
                            pathTerm := ConstrainedPathTerm new.
                            pathTerm ivOffset: ivOffset; constraintClass: (constraintArray at: i)
                        ]
                ]
        ].

    pathTerm moveToDisk.
    pathTerm name: (pathArray at: i); offset: i; addLast: indexObj.
    indexObj addLast: pathTerm.
    pathTerm changeToSegment: GsIndexingSegment
].

" now calculate the children for each path term of the index object "
(1 max: startOffset - 1) to: indexObj size do: [ :i |
    (indexObj at: i) _determineChildren
].

" initialize path term state (whether to update B-trees, dictionary, etc) "
indexObj _setPathTermState.

" add an entry to the NSC's indexed paths list "
self addIndex: indexObj withOffset: 1 nsc: indexObj nscRoot.

self _determineIfConstrained.
%

category: 'Enumerating'
method: IndexList
indexObjectsAndOffsetsDo: aBlock

"Iterates over each (index, offset) pair in the index list, executing the
 two-argument block with the index object and offset as the arguments."
<protected>

1 to: self size by: 2 do: [ :i |
  (self at: i) _isIndexObject
    ifTrue: [ aBlock value: (self at: i) value: (self at: i + 1) ]
]
%

category: 'Enumerating'
method: IndexList
indexObjectsDo: aBlock

"Iterates over each (index, offset) pair in the index list, executing the
 one-argument block with the index object as the argument."
<protected>

1 to: self size by: 2 do: [ :i |
  (self at: i) _isIndexObject
    ifTrue: [ aBlock value: (self at: i) ]
]
%

category: 'Removing'
method: IndexList
removeIndexesInPathTerm: pathTerm for: anNsc

"Removes an entry in the receiver for each index that uses the given path term."
<protected>

" for each index that utilizes the path term "
1 to: pathTerm size do: [ :i |
    " remove the entry in the index list for the NSC "
    self removeIndex: (pathTerm at: i)
        withOffset: pathTerm offset + 1
        for: anNsc
]
%

category: 'Removing'
method: IndexList
removeAllIndexesFor: anNsc

"Removes all indexes from the receiver (it may contain tracking objects)."

<protected>
| indexes |

indexes := Array new.
1 to: self size by: 2 do: [ :i |
  (self at: i + 1) > 0
    ifTrue: [ indexes add: (self at: i); add: (self at: i + 1) ]
].
1 to: indexes size by: 2 do: [ :i |
  self removeIndex: (indexes at: i) withOffset: (indexes at: i + 1) for: anNsc
]
%

category: 'Removing'
method: IndexList
removeIndex: indexObj withOffset: anInteger for: anNsc

"Removes the entry (an entry has 2 slots) containing the given index.  If there
 are no remaining entries, then set the NSC's index list to nil.  Remaining
 entries are moved forward in the list."
<protected>

| i sz |
sz := self size.

i := self _findOffsetFor: indexObj withOffset: anInteger.
i == nil
    ifTrue: [ ^ self ].

" if only one entry, clear the index list "
sz == 2
    ifTrue: [ ^ anNsc _clearIndexList ].

self removeFrom: i to: i + 1 .

self _calculateRootPathTerms.
self _determineIfConstrained.
%

category: 'Testing'
method: IndexList
isConstrained

"Returns the value of the instance variable 'isConstrained'."
<protected>

isConstrained == nil
    ifTrue: [ self _determineIfConstrained ].
^isConstrained
%

category: 'Accessing'
method: IndexList
rootTerms

"Returns the value of the instance variable 'rootTerms'."
<protected>

^rootTerms
%

category: 'Accessing'
method: IndexList
segments

"Returns the value of the instance variable 'segments'."
<protected>

^segments
%

category: 'Updating'
method: IndexList
isConstrained: newValue

"Modifies the value of the instance variable 'isConstrained'."
<protected>

isConstrained := newValue
%

category: 'Updating'
method: IndexList
rootTerms: newValue

"Modifies the value of the instance variable 'rootTerms'."
<protected>

rootTerms := newValue
%

category: 'Updating'
method: IndexList
segments: newValue

"Modifies the value of the instance variable 'segments'."

<protected>

segments := newValue
%

category: 'Initialization'
method: IndexList
initialize

"Initializes the instance with default values."
<protected>

rootTerms := Array new.
isConstrained := false.
segments := SegmentSet new.
%

category: 'Updating'
method: IndexList
_calculateRootPathTerms

"Determines the list of root path terms for the receiver."
<protected>

rootTerms size: 0.

self indexObjectsAndOffsetsDo: [ :indexObj :offset | | pathTerm |
    indexObj size < offset
        ifTrue: [ pathTerm := indexObj lastPathTerm ]
        ifFalse: [ pathTerm := indexObj at: offset ].
    (rootTerms includesIdentical: pathTerm)
        ifFalse: [ rootTerms addLast: pathTerm ]
].
%

category: 'Updating'
method: IndexList
_determineIfConstrained

"Determines if there are any path terms (including child terms) that are not
 constrained."
<protected>

1 to: rootTerms size do: [ :i |
    (rootTerms at: i) isConstrained
        ifFalse: [
            isConstrained := false.
            ^ self
        ]
].
isConstrained := true
%

category: 'Clustering'
method: IndexList
moveToDisk

"Moves the rootTerms to disk, then the receiver."
<protected>

rootTerms moveToDisk.
super moveToDisk
%

category: 'Accessing'
method: IndexList
changeToSegment: aSegment

"Assigns the receiver, its tree of path terms, and the index objects to the
 given segment."
<protected>

| pathTerm |
aSegment == GsIndexingSegment
  ifFalse: [ self _error: #segmentNotSharedDepListSegment ].

self assignToSegment: aSegment.
segments assignToSegment: aSegment.
rootTerms changeToSegment: aSegment.
1 to: rootTerms size do: [ :i |
    pathTerm := rootTerms at: i.
    pathTerm offset == 1
        ifTrue: [ pathTerm changeToSegment: aSegment ]
].
self indexObjectsAndOffsetsDo: [ :indexObj :offset |
    offset == 1
        ifTrue: [ indexObj changeToSegment: aSegment ]
]
%

category: 'Updating'
method: IndexList
addSegment: aSegment logging: aBoolean

"Adds the given segment to the receivers set of segments.  First check if it is
 already in the set.  If not, add it and log the addition in the System redo
 log."

< unprotected primitive: 901>
| logEntry |
(segments _rcIncludes: aSegment)
    ifFalse: [
        self _addSegment: aSegment.
        aBoolean
            ifTrue: [
                logEntry := LogEntry new.
                logEntry receiver: self;
                    selector: #addSegment:logging:;
                    argArray: #[ aSegment, false ].
                System redoLog addLogEntry: logEntry forConflictObject: segments
            ]
    ].
System _disableProtectedMode.
^ true
%

category: 'Testing'
method: IndexList
checkAuthorization

"Determines whether the current user has read authorization on all segments in
 the receiver's segments set.  If not, raise an authorization error."
<protected>

System _addEntireObjectToRcReadSet: segments.
1 to: segments size do: [ :i |
    " this will determine if we can read the segment "
    (segments _at: i) currentUserCanRead
        ifFalse: [
            self
                _error: #rtErrIndexAuthErrSegRead
                args: #[  #'' , segments _at: i ]
        ]
].
%

category: 'Testing'
method: IndexList
includesIndex: indexObj

"Returns whether the given index object is in the receiver."
<protected>

1 to: self size by: 2 do: [ :i |
    indexObj == (self at: i)
        ifTrue: [ ^ true ]
].
^ false
%

category: 'Updating'
method: IndexList
_addSegment: aSegment

"Adds the given segment to the receiver's segments set."
<protected>

segments add: aSegment
%

category: 'Updating Indexes'
method: IndexList
_putAllCommonPathTermsForPathArray: pathArray into: indexObj

"Finds the common path terms that already exist for indexes whose path
 components match pathArray (an Array of Strings).  Add those path terms to
 the end of given indexObj."
<protected>

| firstUnique somePathTerms j minSize sz |
firstUnique := 0.
sz := pathArray size.
1 to: self size by: 2 do: [ :i |
    somePathTerms := self at: i.
    j := self at: i + 1.
    " only consider index objects for which the NSC is the root "
    j == 1
        ifTrue: [
            minSize := somePathTerms size min: sz.

            [ j <= minSize _and:
            [ (somePathTerms at: j) name = (pathArray at: j) ] ] whileTrue: [
                j > firstUnique
                   ifTrue: [
                       firstUnique := j.
                       indexObj addLast: (somePathTerms at: j)
                   ].
                j := j + 1
            ]
        ]
].
%

category: 'Accessing'
method: IndexList
at: anIndex

"Protected version of Object | at:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [^ self _errorNonIntegerIndex: anIndex].
self _primitiveFailed: #at: .
self _uncontinuableError
%

category: 'Accessing'
method: IndexList
instVarAt: anIndex

"Protected version of Object | instVarAt:."

<protected primitive: 811"with rcvrClass">

| numInstVars |
(anIndex _isInteger)
ifTrue:
[
   numInstVars:= self class allInstVarNames size.
   (numInstVars == 0)
   ifTrue:
      [ ^ self _error: #rtErrNoInstVars].
   ((anIndex >= 1) & (anIndex <= numInstVars))
   ifFalse: "out of bounds"
      [ ^ self _errorIndexOutOfRange: anIndex]
]
ifFalse: "anIndex not an integer"
   [ ^ self _errorNonIntegerIndex: anIndex
].

self _primitiveFailed: #instVarAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IndexList
_at: anIndex

"Protected version of Object | _at:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_at: .
self _uncontinuableError
%

category: 'Accessing'
method: IndexList
_basicAt: anIndex

"Protected version of Object | _basicAt:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_basicAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IndexList
_primitiveAt: anIndex

"Protected version of Object | _primitiveAt:."

<protected primitive: 89>

(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_primitiveAt: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
at: anIndex put: aValue

"Protected version of Object | at:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue: [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #at:put: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
instVarAt: anIndex put: aValue

"Protected version of Object | instVarAt:put:."

<protected primitive: 804"with rcvrClass">
(self class allInstVarNames size == 0)
  ifTrue: [ ^ self _error: #rtErrNoInstVars].
(anIndex _isInteger)
  ifTrue: [((anIndex < 1) | (anIndex > self class allInstVarNames size))
        ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]] "out of bounds"
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #instVarAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
size: anInteger

"Protected version of Object | size:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
_basicAt: anIndex put: aValue

"Protected version of Object | _basicAt:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_basicAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
_at: anIndex put: aValue

"Protected version of Object | _at:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_at:put: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
_basicSize: anInteger

"Protected version of Object | _basicSize:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: IndexList
_primitiveAt: anIndex put: aValue

"Protected version of Object | _primitiveAt:put:."

<protected primitive: 281>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
((anIndex < 1) | (anIndex > (self _primitiveSize + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_primitiveAt:put: .
self _uncontinuableError
%

category: 'Reduced Conflict Support'
method: IndexList
_rcAt: anIndex

"Protected version of Object | _rcAt:."

<protected primitive: 92>
self _primitiveFailed: #_rcAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IndexList
basicAt: anIndex

"Protected version of Object | basicAt:."
<protected>

^self _basicAt: anIndex
%

category: 'Updating'
method: IndexList
basicAt: anIndex put: aValue

"Protected version of Object | basicAt:put:."
<protected>

^self _basicAt: anIndex put: aValue
%

category: 'Copying'
method: IndexList
copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex

"Protected version of Array | copyFrom:to:into:startingAt:."

<protected primitive: 808"with rcvrClass">

^ super copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex
%

category: 'Copying'
method: IndexList
removeFrom: startIndex to: stopIndex

"Protected version of Array | removeFrom:to:."

<protected primitive: 801>

^ super removeFrom: startIndex to: stopIndex
%
category: 'Removing'
method: IndexList
_deleteNoShrinkFrom: startIndex to: stopIndex

"Protected version of Array | _deleteNoShrinkFrom:to:."

<protected primitive: 802 "with rcvrClass">
^ super _deleteNoShrinkFrom: startIndex to: stopIndex
%


category: 'Copying'
method: IndexList
_insertAt: destOffset
from: anArray
fromStart: startOffset
fromEnd: endOffset
numToMoveDown: numToMove

"Protected version of Array | #_insertAt:from:fromStart:fromEnd:numToMoveDown:."

<protected primitive: 805"with rcvrClass">
^ super _insertAt: destOffset from: anArray fromStart: startOffset
            fromEnd: endOffset numToMoveDown: numToMove
%

category: 'Copying'
method: IndexList
insertAll: aSeqColl at: anIndex

"Protected version of Array | insertAll:at:."

<protected primitive: 806"with rcvrClass">

^ super insertAll: aSeqColl at: anIndex
%

category: 'Adding'
method: IndexList
add: newObject

"Protected version of Array | add:."

<protected primitive: 203>
self _primitiveFailed: #add: .
self _uncontinuableError
%

category: 'Adding'
method: IndexList
addLast: newObject

"Protected version of Array | addLast:."

<protected primitive: 203>
self _primitiveFailed: #addLast: .
self _uncontinuableError
%

category: 'Adding'
method: IndexList
_addAllFromNsc: aBag

"Protected version of Array | _addAllFromNsc:."

<protected primitive: 202>
self _primitiveFailed: #_addAllFromNsc: .
self _uncontinuableError
%

category: 'Searching'
method: IndexList
_numberOfCommonPathTermsForPathArray: pathArray

"Returns the number of path terms that already exist for indexes whose path
 components match pathArray (an Array of Strings)."
<protected>

| firstUnique somePathTerms j continue count k |
firstUnique := 0.
count := 0.
k := 1.
1 to: self size by: 2 do: [ :i |
  somePathTerms := self at: i.
  j := self at: i + 1.
  j > 0
    ifTrue: [
      continue := true.

      [ k <= pathArray size _and: [ j <= somePathTerms size _and:
      [ (somePathTerms at: j) name = (pathArray at: k) ] ] ] whileTrue: [
        j > firstUnique
         ifTrue: [
           firstUnique := j.
           count := count + 1
         ].
        j := j + 1.
        k := k + 1.
      ]
    ]
].
^ count
%

category: 'Reduced Conflict Support'
method: IndexList
_selectiveAbort

"The only thing that should be selectively aborted for an index list is
 the segments set."

< unprotected >

segments _selectiveAbort .
%

category: 'Testing'
method: IndexList
_checkAllIndexAuthorizations

"Check whether the user has read permission for all indexes on the receiver."
<protected>

self checkAuthorization.
1 to: self size by: 2 do: [ :i |
  (self at: i) _isIndexObject
    ifTrue: [ (self at: i) checkAuthorization ]
]
%

category: 'Testing'
method: IndexList
_hasExplicitIndex

"Returns whether the receiver has an explicitly created index for it."
<protected>

2 to: self size by: 2 do: [ :i |
  (self at: i) == 1
    ifTrue: [ ^ true ]
].
^ false
%

category: 'Updating Indexes'
method: IndexList
_emptySegmentSets

"Removes all segments from the receiver's segment set and from the
 root terms and their children's recursively."
<protected>

segments removeAll: segments.
1 to: rootTerms size do: [ :i |
  (rootTerms at: i) _emptySegmentSets
]
%

category: 'Statistics'
method: IndexList
_statisticsInto: dict

"Put statistical information into the given dictionary."
<protected>

| pages indexObj pathTerm |
pages := IdentitySet new.

System _addEntireObjectToRcReadSet: self.
pages add: self page.

System _addEntireObjectToRcReadSet: rootTerms.
pages add: rootTerms page.

System _addEntireObjectToRcReadSet: segments.
pages add: segments page.


1 to: self size by: 2 do: [ :i |
  indexObj := self at: i.
  indexObj _isIndexObject
    ifTrue: [

      System _addEntireObjectToRcReadSet: indexObj.
      indexObj _statisticsInto: dict.

      pages add: indexObj page.

      1 to: indexObj size do: [ :j |
        pathTerm := indexObj at: j.

        System _addEntireObjectToRcReadSet: pathTerm.
        pages add: pathTerm page.

        System _addEntireObjectToRcReadSet: pathTerm segments.
        pages add: pathTerm segments page.

        System _addEntireObjectToRcReadSet: pathTerm children.
        pages add: pathTerm children page.
      ]
    ]
].
dict at: #IndexObjectPages put: pages
%

category: 'Accessing'
method: IndexList
_classOfElements

"Returns the kind of objects stored in the NSC that has the receiver as its
 _indexedPaths.  This can only be determined if there is an equality index on
 the elements of the NSC.  In that case, the index object holds the value of the
 lastElementClass instance variable; otherwise returns the class Object."

<protected>

| pathTerm |
1 to: rootTerms size do: [ :i |
  pathTerm := rootTerms at: i.
  pathTerm indicatesIndexOnNscElements
    ifTrue: [
      1 to: pathTerm size do: [ :i |
        ( (pathTerm at: i) isRangeEqualityIndex _and:
        [ (pathTerm at: i) size == pathTerm offset ] )
          ifTrue: [ ^ (pathTerm at: i) lastElementClass ]
      ]
    ].
].
^ Object
%

category: 'Updating'
method: IndexList
addTracker: tracker

"Adds the tracking object to the receiver."

<protected>
" do not allow duplicate entries "
1 to: self size by: 2 do: [ :i |
    " if entry already exists, just returns "
  tracker == (self at: i)
    ifTrue: [ ^ self ]
].

" add the entry to the end of the index list "
self addLast: tracker.
self addLast: 0.
%

category: 'Removing'
method: IndexList
removeTracker: tracker for: anNsc

"Removes the tracking object from the receiver."

<protected>
| i |

i := self _findOffsetFor: tracker.
i == nil
  ifTrue: [ ^ self ].

" if only one entry, clear the index list "
self size == 2
  ifTrue: [ ^ anNsc _clearIndexList ].

self removeFrom: i to: i + 1 .
%

category: 'Updating'
method: IndexList
hasTrackingObjects

"Returns whether any tracking objects are contained in the receiver."

<protected>
2 to: self size by: 2 do: [ :i |
  (self at: i) > 0
    ifFalse: [ ^ true ]
].
^ false
%

category: 'Testing'
method: IndexList
_hasASetValuedTerm
  "Answers true if the receiver contains a set-valued term."
<protected>
self indexObjectsDo: [:iObj |
  iObj hasSetValuedTerm ifTrue: [ ^ true ] ].
^ false
%
unprotectmethods

